/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.cvc3generator;

import ca.mcscert.jtet.expression.*;
import ca.mcscert.jtet.expression.ExpressionGeneratorSpy;
import ca.mcscert.jtet.parsers.MatlabParser;
import ca.mcscert.jtet.tablularexpression.HierarchicalGridBreadthFirstCheckerGenerator;
import ca.mcscert.jtet.tablularexpression.Cell;


import java.util.ArrayList;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.List;

import org.apache.commons.lang3.StringUtils;

/**
 *
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
final public class HierarchicalGridCVC3Generator implements HierarchicalGridBreadthFirstCheckerGenerator {

    public HierarchicalGridCVC3Generator(VariableCollection variableDefinitions, int queryCountStart) {
        m_variableDefinitions = variableDefinitions;
        m_currentTCC = queryCountStart;
    }

    private String generateQueryPrefix() {
        if (!m_parentCellsCVC3Code.isEmpty()){
            String output = "(";
            output += StringUtils.join(m_parentCellsCVC3Code, " AND ");
            output += ") => ";
            return output;
        }
        return "";
    }

    private void outputCells() {
        // First is disjoint testing.
        m_output += "ECHO \"begin" + m_currentTCC + "\";\n";
        m_output += "PUSH;\n";

        String query = "QUERY ";
        query += generateQueryPrefix();
        query += "(" + StringUtils.join(m_currentDisjointQueries, " AND\n");
        query += ");";
        m_queries.add(query);
        m_output += query + "\n";

        m_output += "POP;\n";
        m_output += "ECHO \"end" + m_currentTCC + "\";\n";

        // Finished first TCC, increment counter.
        m_currentTCC++;

        // Next is complete.
        m_output += "ECHO \"begin" + m_currentTCC + "\";\n";
        m_output += "PUSH;\n";

        query = "QUERY ";
        query += generateQueryPrefix();
        query += "(" + StringUtils.join(m_currentCompleteQueries, " OR ");
        query += ");";
        m_queries.add(query);
        m_output += query + "\n";

        m_output += "POP;\n";
        m_output += "ECHO \"end" + m_currentTCC + "\";\n";

        // Finished other TCC, increment counter.
        m_currentTCC++;

        m_currentlyRunning = false;

        // And finally clear the running list of current queries.
        m_currentCompleteQueries.clear();
        m_currentDisjointQueries.clear();
    }

    @Override
    public void descendIntoGridFromCell(Cell cell) {
        if (m_currentlyRunning) {
            outputCells();
        }
        m_parentCellsCVC3Code.addFirst(MatlabParser.parseMatlabCode(m_variableDefinitions, cell.contents()).getCheckerOutput(m_CVC3ExpressionGenerator, BooleanVariableType.Type));
        m_currentlyRunning = true;
    }

    @Override
    public void ascendFromGrid() {
        if (m_currentlyRunning) {
            outputCells();
        }
        m_parentCellsCVC3Code.removeFirst();
    }

    @Override
    public void handleEdgeCell(Cell inputCell) {
        handleLeafCell(inputCell);
    }

    @Override
    public void handleLeafCell(Cell cell) {
        // First get the appropriate CVC3 code from matlab.  Capture generation for further processing.
        ExpressionGeneratorSpy spy = new ExpressionGeneratorSpy(m_CVC3ExpressionGenerator);
        String newCellCode = MatlabParser.parseMatlabCode(m_variableDefinitions, cell.contents()).getCheckerOutput(spy, BooleanVariableType.Type);

        // Next, find all the fixed point expressions and handle them.
        for(ExpressionGeneratorSpy.SpiedBinaryOp op : spy.getBinaryOps()) {
            if (op.usedType instanceof FixedPointVariableType) {
                generateOverflowTCC(op);
            }
        }

        // Next form the disjoint queries and add to the list.
        for(String otherCellCode : m_currentCompleteQueries) {
            String newDisjointQuery = "NOT ( " + otherCellCode  + " AND " + newCellCode + " )";
            m_currentDisjointQueries.add(newDisjointQuery);
        }

        // Lastly add the cell to the list for the complete query
        m_currentCompleteQueries.add(newCellCode);
    }

    private void generateOverflowTCC(ExpressionGeneratorSpy.SpiedBinaryOp binaryOp) {
        // Comparisons never overflow, so ignore them.
        switch (binaryOp.op) {
            case GreaterThen:
            case GreaterThenEqual:
            case LessThen:
            case LessThenEqual:
            case Equals:
            case NotEquals:
                return;
        }
        FixedPointVariableType type = (FixedPointVariableType) binaryOp.usedType;
        if (!type.isSigned() && binaryOp.op == BinaryOperation.Division) {
            return ; // Skip this generation.  Unsigned division requires no checks, since it never over/under flows.
        }

        m_output += "ECHO \"begin" + m_currentTCC + "\";\n";
        m_output += "PUSH;\n";

        String query = null;
        if (((FixedPointVariableType) binaryOp.usedType).isSigned()) {
            switch (binaryOp.op) {
                case Addition:
                    // Verify the sign bit doesn't change as appropriate.  Only on the sign bits being equal can an error occur, and only if the output sign differs.
                    query = "QUERY ((x[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin0 AND y[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin0) " +
                                "=> BVPLUS(" + type.digits() + ",x,y)[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin0) AND " +
                            "((x[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin1 AND y[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin1) " +
                                "=> BVPLUS(" + type.digits() + ",x,y)[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin1);";
                    break;
                case Minus:
                    // Verify the sign bit doesn't change as appropriate.  Only on the sign bits being equal can an error occur, and only if the output sign differs.
                    query = "QUERY ((x[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin0 AND y[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin1) " +
                            "=> BVPLUS(" + type.digits() + ",x,y)[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin0) AND " +
                            "((x[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin1 AND y[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin0) " +
                            "=> BVPLUS(" + type.digits() + ",x,y)[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = 0bin1);";
                    break;
                case Multiplication:
                    // Same as unsigned, sign extend and verify the extra bits don't change to the wrong value.  Same signs is all 0, different signs are all 1.
                    query = "QUERY ((x[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] = y[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "]) => BVMULT(" +
                            (type.digits() * 2) + ",SX(" + binaryOp.lhsExp + "," + (type.digits() * 2) + "),SX(" + binaryOp.rhsExp + "," + (type.digits() * 2) + "))[" +
                            (type.digits() * 2 - 1) + ":" + (type.digits() - 1) + "] = 0bin" + StringUtils.repeat('0', type.digits() + 1) + ") AND " +
                            "((x[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "] /= y[" + (type.digits() - 1) + ":" + (type.digits() - 1) + "]) => BVMULT(" +
                            (type.digits() * 2) + ",SX(" + binaryOp.lhsExp + "," + (type.digits() * 2) + "),SX(" + binaryOp.rhsExp + "," + (type.digits() * 2) + "))[" +
                            (type.digits() * 2 - 1) + ":" + (type.digits() - 1) + "] = 0bin" + StringUtils.repeat('1', type.digits() + 1) + ");";
                    break;
                case Division:
                    // Overflow only happens when the smallest negative is multiplied by -1, as that can't be represented in the positive.
                    query = "QUERY NOT (x = 0bin1" + StringUtils.repeat('0', type.digits() - 1) + " AND y = 0bin" + StringUtils.repeat('1', type.digits()) + ");";
                    break;
            }
        } else {
            switch (binaryOp.op) {
            case Addition:
                // Extend the addition and verify the extra bit remains unchanged.
                query = "QUERY BVPLUS(" + (type.digits() + 1) + "," + binaryOp.lhsExp + "," + binaryOp.rhsExp + ")[" +
                        (type.digits() + 1) + ":" + type.digits() + "] = 0bin0;";
                break;
            case Minus:
                // Test rhs <= lhs, which guarantees this case as (lhs >= rhs => lhs - rhs > 0).
                query = "QUERY BVLE(" + binaryOp.rhsExp + "," + binaryOp.lhsExp + ");";
                break;
            case Multiplication:
                // Extend the addition and test extra bits remain zero, like addition.
                query = "QUERY BVMULT(" + (type.digits() * 2) + "," + binaryOp.lhsExp + "," + binaryOp.rhsExp + ")[" +
                        (type.digits() * 2 - 1) + ":" + type.digits() + "] = 0bin" + StringUtils.repeat('0', type.digits()) + ";";
                break;
            case Division:
                // Unsigned never overflows, as the result never grows.  We skip any generation, so we should never come here.
                throw new IllegalStateException("No check necessary for unsigned division!  Should never get here!");
            }
        }
        m_queries.add(query);
        m_output += query + "\n";

        m_output += "POP;\n";
        m_output += "ECHO \"end" + m_currentTCC + "\";\n";
        m_currentTCC++; // For now, just screw up the TCC counting.
    }

    @Override
    public String getFinalString() {
        if (m_currentlyRunning) {
            outputCells();
        }
        return m_output;
    }

    public List<String> getFinalQueries() {
        if (m_currentlyRunning) {
            outputCells();
        }
        return m_queries;
    }

    Deque<String> m_parentCellsCVC3Code = new ArrayDeque<String>();

    List<String> m_currentDisjointQueries = new ArrayList<String>();
    List<String> m_currentCompleteQueries = new ArrayList<String>();

    boolean m_currentlyRunning = true;
    int m_currentTCC = 1;

    String m_output = "";
    List<String> m_queries = new ArrayList<String>();

    // To fix properly.
    VariableCollection m_variableDefinitions;
    CVC3ExpressionGenerator m_CVC3ExpressionGenerator = new CVC3ExpressionGenerator();
}
