/*
 * Copyright (C) 2014 Yanjun Jiang <jiangy76@mcmaster.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.eventbgenerator;

import ca.mcscert.jtet.tablularexpression.*;

import java.io.IOException;
import java.util.List;

/**
 *
 * @author Yanjun Jiang <jiangy76@mcmaster.ca>
 */
final public class EventBTableGenerator {

    public EventBTableGenerator(Table table, RefinementMode refineMode) {
        m_table = table;
        m_refineMode = refineMode;
    }

    public void generateEventBXml(String folderPath) throws IOException {
        final String versionEncodingXml = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n";
        final String machineFileCfgPreXml = "<org.eventb.core.machineFile org.eventb.core.configuration=\"org.eventb.core.fwd;de.prob.units.mchBase\" version=\"5\">\n";
        final String machineFileCfgPostXml = "</org.eventb.core.machineFile>";

        int outputLayerNum = calculateOutputLayerNum();
        for (int i = 0; i < outputLayerNum; i++) {

            //Step 1: Append the Pre Xml
            String output = "";
            output += versionEncodingXml;
            output += machineFileCfgPreXml;

            //Step 2: Generate the Variables & Invariants Xml
            EventBVariablesDeclarationGenerator eventBVariablesDeclarationGenerator = new EventBVariablesDeclarationGenerator();
            output += eventBVariablesDeclarationGenerator.GenerateVariablesDeclaration(m_table.getVariables());

            //Step 3: Generate the Events Xml
            HierarchicalGridEventBGenerator hierarchicalGridEventBGenerator;
            int refineLayerNo = (RefinementMode.WithoutRefinement == m_refineMode) ? Integer.MAX_VALUE : i;//No Refinement mode layerCount is set to infinity so that handleLeafCell is always executed.
            hierarchicalGridEventBGenerator = new HierarchicalGridEventBGenerator(m_refineMode, refineLayerNo, m_table);
            output += m_table.walk(hierarchicalGridEventBGenerator);

            //Step 4: Generate the file Refine Relation Xml
            if (hierarchicalGridEventBGenerator.isRefineRelationNecessary()) {
                output += generateFileRefineRelation(refineLayerNo);
            }

            //Step 5: Append the Post Xml
            output += machineFileCfgPostXml;

            //Step 6: Write out EventB code into file in the project folder
            EventBFileWriter.writeEventBBumFile(getFullFileName(i, folderPath) ,output);
        }
    }

    String getFullFileName(int no, String folderPath) {
        String fullName = folderPath;

        if (m_refineMode == RefinementMode.WithRefinement) {
            fullName += generateFileName(no) + ".bum";
        } else {
            fullName += m_table.getTableName()  + ".bum";
        }

        return fullName;
    }

    private int calculateOutputLayerNum() {
        if (RefinementMode.WithoutRefinement  == m_refineMode) {
            return 1;
        }
        else {
            return calculateMaxRefineLayerNum(m_table.getLeftGrid());
        }
    }

    //m_maxRefineLayerNum is also the max Recursion Depth
    private int calculateMaxRefineLayerNum(SubHierarchyFetcher grid) {
        int maxRecursionDepth = 0;
        List<HierarchicalCell> cells = grid.getSubHierarchy();

        for (HierarchicalCell cell : cells) {
            int subGridMaxRefineLayerNum = calculateMaxRefineLayerNum(cell) + 1;
            if (subGridMaxRefineLayerNum > maxRecursionDepth){
                maxRecursionDepth = subGridMaxRefineLayerNum;
            }
        }

        return maxRecursionDepth;
    }

    private String generateFileName(int fileNo) {
        return m_table.getTableName() + "_M" + fileNo;
    }

    private String generateFileRefineRelation(int refinerFileNo) {
        int refinedFileNo = refinerFileNo - 1;
        return "<org.eventb.core.refinesMachine name=\"" + generateFileName(refinerFileNo) + "\" org.eventb.core.target=\"" + generateFileName(refinedFileNo) + "\"/>\n";
    }

    Table m_table;
    RefinementMode m_refineMode;
}
