/*
 * Copyright (C) 2014 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package ca.mcscert.jtet.expression;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * This inserts a spy between a given ExpressionGenerator and the caller, recording all operations.
 * <p>
 * This is useful if it is wanted to verify that all operations satisfy some property, for instance that all fixed point
 * operations do not under/overflow.
 * @author Matthew Dawson
 */
public class ExpressionGeneratorSpy implements ExpressionGenerator {
    /**
     * Constructs the spy, using the given ExpressionGenerator to generate the actual expressions.
     * @param expressionGenerator The ExpressionGenerator to use to actually generate the expressions.
     */
    public ExpressionGeneratorSpy(ExpressionGenerator expressionGenerator) {
        this.expressionGenerator = expressionGenerator;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String GenerateUnaryOperation(UnaryOperation op, String expression, VariableType usedType) {
        m_spiedUnaryOps.add(new SpiedUnaryOp(op, expression, usedType));
        return expressionGenerator.GenerateUnaryOperation(op, expression, usedType);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String GenerateBinaryOperation(BinaryOperation op, String lhsExp, String rhsExp, VariableType usedType) {
        m_spiedBinaryOps.add(new SpiedBinaryOp(lhsExp, op, rhsExp, usedType));
        return expressionGenerator.GenerateBinaryOperation(op, lhsExp, rhsExp, usedType);
    }

    /**
     * {@inheritDoc}
     * <p>
     * This currently remains unrecorded.
     */
    @Override
    public String GenerateLiteralValue(String value, VariableType requestedType) {
        return expressionGenerator.GenerateLiteralValue(value, requestedType);
    }

    /**
     * Returns an unmodifiable list of all unary operations that have occurred so far.
     * @return All unary operations seen
     */
    public List<SpiedUnaryOp> getUnaryOps() {
        return Collections.unmodifiableList(m_spiedUnaryOps);
    }

    /**
     * Returns an unmodifiable list of all binary operations that have occurred so far.
     * @return All binary operations seen
     */
    public List<SpiedBinaryOp> getBinaryOps() {
        return m_spiedBinaryOps;
    }

    /**
     * This class contains the details of a unary operation: The actual operation, the input expression, and the output type.
     */
    public static class SpiedUnaryOp {
        /**
         * The operator for this operation.
         */
        public final UnaryOperation op;
        /**
         * The input expression to the operator.
         */
        public final String expression;
        /**
         * The output type of the expression.
         */
        public final VariableType usedType;

        /**
         * Constructs a SpiedUnaryOp with the details filled in.
         * @param op The operator for this operation
         * @param expression The input expression to the operator
         * @param usedType The output type of the expression
         */
        private SpiedUnaryOp(UnaryOperation op, String expression, VariableType usedType) {
            this.op = op;
            this.expression = expression;
            this.usedType = usedType;
        }
    }

    /**
     * This class contains the details of a binary operation: The actual operation, the input expressions, and the output type.
     */
    public static class SpiedBinaryOp {
        /**
         * The operator for this operation.
         */
        public final BinaryOperation op;
        /**
         * The expression on the left hand side of the operator.
         */
        public final String lhsExp;
        /**
         * The expression on the right hand side of the operator.
         */
        public final String rhsExp;
        /**
         * The output type of the expression.
         */
        public final VariableType usedType;

        /**
         * Constructs a SpiedBinaryOp with the details filled in.
         * @param lhsExp The expression on the left hand side of the operator
         * @param op The operator for this operation
         * @param rhsExp The expression on the right hand side of the operator
         * @param usedType The output type of the expression
         */
        private SpiedBinaryOp(String lhsExp, BinaryOperation op, String rhsExp, VariableType usedType) {
            this.lhsExp = lhsExp;
            this.op = op;
            this.rhsExp = rhsExp;
            this.usedType = usedType;
        }
    }

    private final ExpressionGenerator expressionGenerator;
    private final List<SpiedUnaryOp> m_spiedUnaryOps = new ArrayList<SpiedUnaryOp>();

    private final List<SpiedBinaryOp> m_spiedBinaryOps = new ArrayList<SpiedBinaryOp>();
}
