/*
 * Copyright (C) 2014 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.parsers;

import ca.mcscert.jtet.expression.BooleanVariableType;
import ca.mcscert.jtet.expression.EnumerationVariableType;
import ca.mcscert.jtet.expression.FixedPointVariableType;
import ca.mcscert.jtet.expression.PartialVariableCollection;
import ca.mcscert.jtet.expression.RealVariableType;
import ca.mcscert.jtet.expression.Variable;
import ca.mcscert.jtet.expression.VariableCollection;
import ca.mcscert.jtet.parsers.antlr.VariableParserParser.Bool_typeContext;
import ca.mcscert.jtet.parsers.antlr.VariableParserParser.DefaulttypeContext;
import ca.mcscert.jtet.parsers.antlr.VariableParserParser.Fixed_typeContext;
import ca.mcscert.jtet.parsers.antlr.VariableParserParser.Real_typeContext;
import ca.mcscert.jtet.parsers.antlr.VariableParserParser.VarContext;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import ca.mcscert.jtet.expression.VariableType;
import ca.mcscert.jtet.parsers.antlr.VariableParserLexer;
import ca.mcscert.jtet.parsers.antlr.VariableParserParser;
import ca.mcscert.jtet.parsers.antlr.VariableParserParserBaseListener;
import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.misc.NotNull;
import org.antlr.v4.runtime.tree.ParseTree;
import org.antlr.v4.runtime.tree.ParseTreeWalker;
import org.antlr.v4.runtime.tree.TerminalNode;

/**
 *
 * @author Matthew Dawson
 */
final public class VariableParser {

    public PartialVariableCollection parseVariables(String variables) {
        // create a CharStream that reads from standard input
        ANTLRInputStream input = new ANTLRInputStream(variables) {

            @Override
            public int LA(int i) {
                return Character.toLowerCase(super.LA(i));
            }

        };

        // create a lexer that feeds off of input CharStream
        VariableParserLexer lexer = new VariableParserLexer(input);

        // create a buffer of tokens pulled from the lexer
        CommonTokenStream tokens = new CommonTokenStream(lexer);

        // create a parser that feeds off the tokens buffer
        VariableParserParser parser = new VariableParserParser(tokens);

        ParseTree tree = parser.varlist();
        ParseTreeWalker walker = new ParseTreeWalker();
        VariablesParserToVariables listener = new VariablesParserToVariables();
        walker.walk(listener, tree);

        return listener.m_variableCollection;
    }

    public void addCustomType(String typeName, VariableType type) {
        m_customTypes.put(typeName, type);
    }

    private final Map<String, VariableType> m_customTypes = new HashMap<String, VariableType>();

    private class VariablesParserToVariables extends VariableParserParserBaseListener {
        private String currentName;
        private VariableType currentType;
        private limitingExpressionData currentLimitingExpression;

        @Override
        public void enterVar(VarContext ctx) {
            currentName = ctx.ID().getText();
        }

        @Override
        public void enterDefaulttype(DefaulttypeContext ctx) {
            currentType = new RealVariableType();
        }

        @Override
        public void enterReal_type(Real_typeContext ctx) {
            currentType = new RealVariableType();
        }

        @Override
        public void enterBool_type(Bool_typeContext ctx) {
            currentType = new BooleanVariableType();
        }

        @Override
        public void enterUnknown_type(@NotNull VariableParserParser.Unknown_typeContext ctx) {
            currentType = m_customTypes.get(ctx.UNKNOWN_TYPE().getText());
            if (currentType == null) { // Assume this unknown.  For now, pretend a real is wanted.
                currentType = new RealVariableType();
            } else if (currentType instanceof EnumerationVariableType) {
                m_usedEnumerationTypes.add((EnumerationVariableType)currentType);
            }
        }

        @Override
        public void enterSubtype_var(@NotNull VariableParserParser.Subtype_varContext ctx) {
            currentLimitingExpression = new limitingExpressionData();
            currentLimitingExpression.variableName = ctx.ID().getText();
        }

        @Override
        public void enterPvs_expr(@NotNull VariableParserParser.Pvs_exprContext ctx) {
            currentLimitingExpression.expression = ctx.PVS_EXPR().getText();
        }

        @Override
        public void enterFixed_type(Fixed_typeContext ctx) {
            boolean signed = (ctx.NAT(0).getText().equals("0"))?false:true;
            int digits = Integer.parseInt(ctx.NAT(1).getText());
            int fractionPart = 0;
            TerminalNode fractionNode;
            if ((fractionNode = ctx.NAT(2)) != null) {
                fractionPart = Integer.parseInt(fractionNode.getText());
            }
            currentType = new FixedPointVariableType(signed, digits, fractionPart);
        }

        @Override
        public void exitVar(VarContext ctx) {
            m_variables.put(currentName, new Variable(currentName, currentType));
            if (currentLimitingExpression != null) {
                m_variableLimitingExpressions.put(currentName, currentLimitingExpression);
            }
            currentName = null;
            currentType = null;
            currentLimitingExpression = null;
        }

        @Override
        public void exitVarlist(@NotNull VariableParserParser.VarlistContext ctx) {
            for(Map.Entry<String, limitingExpressionData> limitingExpression : m_variableLimitingExpressions.entrySet()) {
                String subtypeVar = limitingExpression.getValue().variableName;
                String subtypeExpression = limitingExpression.getValue().expression;
                if (m_variables.containsKey(limitingExpression.getValue().variableName)) {
                    throw new IllegalStateException("Reused variable name for subtype declaration!");
                }
                // Get the variable we are working on, and use it for the variable used in the subtype expression.  This
                // effectively means the expression can just be emitted into the verification system to limit it.
                m_variables.put(subtypeVar, m_variables.get(limitingExpression.getKey()));

                VariableCollection variableCollection = new VariableCollection(new PartialVariableCollection(m_variables, m_usedEnumerationTypes));
                m_variables.get(limitingExpression.getKey()).type().setSubtypePredicate(PVSSimpleParser.parsePVSCode(variableCollection, subtypeExpression));

                // Undo the variables hack, to avoid having unused variables bleed into the verification scripts!
                m_variables.remove(subtypeVar);
            }

            m_variableCollection = new PartialVariableCollection(m_variables, m_usedEnumerationTypes);
        }

        private Map<String, Variable>  m_variables = new LinkedHashMap<String, Variable>();
        private Set<EnumerationVariableType> m_usedEnumerationTypes = new HashSet<EnumerationVariableType>();
        public PartialVariableCollection m_variableCollection;

        private Map<String, limitingExpressionData> m_variableLimitingExpressions = new HashMap<String, limitingExpressionData>();
    }
    private static class limitingExpressionData {
        public String expression;
        public String variableName;
    }
}
