/*
 * Copyright (C) 2014 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.tablularexpression;

import ca.mcscert.jtet.expression.Variable;
import ca.mcscert.jtet.expression.VariableCollection;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
public class Table {
    public Table(String tableName, VariableCollection variables) {
        this.tableName = tableName;
        this.m_variables = variables;
    }

    public void verifyTable() throws IllegalTableSetup {
        if (m_variables.getOutputVariables().size() != variableOutputs.size()) {
            throw new IllegalTableSetup("The number of output variables and the number of outputs disagree!");
        }

        int leftCount = leftGrid.countTotalCases();
        int topCount = topGrid.countTotalCases();
        for (final Map.Entry<Variable, TwoDimensionalGrid> variableOutput : variableOutputs.entrySet()) {
            TwoDimensionalGrid output = variableOutput.getValue();
            if (output.sizeX() != topCount || output.sizeY() != leftCount) {
                throw new IllegalTableSetup("There are a different number of outputs then conditions for variable " + variableOutput.getKey().name() + "!");
            }
        }
    }

    public String getTableName() {
        return tableName;
    }

    public HierarchicalGrid getLeftGrid() {
        return leftGrid;
    }

    public HierarchicalGrid getTopGrid() {
        return topGrid;
    }

    public Map<Variable, TwoDimensionalGrid> getVariableOutputs() {
        return variableOutputs;
    }

    public VariableCollection getVariables() {
        return m_variables;
    }

    private int HandleGrid(SubHierarchyFetcher grid, HierarchicalGridDepthFirstCheckerGenerator generator, int outputCellIndex) {
        List<HierarchicalCell> cells = grid.getSubHierarchy();
        for (int i = 0; i < cells.size(); ++i) {
            HierarchicalCell cell = cells.get(i);

            if (cell.getSubHierarchy().isEmpty()) {
                Map<Variable, Cell> outputCells = new LinkedHashMap<Variable, Cell>();
                for(Variable outputVariable : m_variables.getOutputVariables().values()) {
                    outputCells.put(outputVariable, variableOutputs.get(outputVariable).get(0, outputCellIndex));
                }
                generator.handleLeafCell(cell, outputCells);
                outputCellIndex++;
            } else {
                generator.handleEdgeCell(cell);
                generator.descendIntoGridFromCell(cell);
                outputCellIndex = HandleGrid(cell, generator, outputCellIndex);
                generator.ascendFromGrid();
            }
        }
        return outputCellIndex;
    }

    public String walk(HierarchicalGridDepthFirstCheckerGenerator generator) {
        HandleGrid(leftGrid, generator, 0);
        return generator.getFinalString();
    }

    private static void HandleGrid(SubHierarchyFetcher grid, HierarchicalGridBreadthFirstCheckerGenerator generator) {
        List<HierarchicalCell> cells = grid.getSubHierarchy();
        for (int i = 0; i < cells.size(); ++i) {
            HierarchicalCell cell = cells.get(i);

            if (cell.getSubHierarchy().isEmpty()) {
                generator.handleLeafCell(cell);
            } else {
                generator.handleEdgeCell(cell);
            }
        }

        for (int i = 0; i < cells.size(); ++i) {
            HierarchicalCell cell = cells.get(i);

            if (!cell.getSubHierarchy().isEmpty()) {
                generator.descendIntoGridFromCell(cell);
                HandleGrid(cell, generator);
                generator.ascendFromGrid();
            }
        }
    }

    public String walk(HierarchicalGridBreadthFirstCheckerGenerator generator) {
        HandleGrid(leftGrid, generator);
        return generator.getFinalString();
    }

    private String tableName;
    private VariableCollection m_variables;
    private final HierarchicalGrid leftGrid = new HierarchicalGrid();
    private final HierarchicalGrid topGrid = new HierarchicalGrid();
    private final Map<Variable, TwoDimensionalGrid> variableOutputs = new HashMap<Variable, TwoDimensionalGrid>();
}
