/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.cvc3generator;

import ca.mcscert.jtet.expression.VariableCollection;
import ca.mcscert.jtet.parsers.VariableParser;
import ca.mcscert.jtet.tablularexpression.HierarchicalCell;
import ca.mcscert.jtet.tablularexpression.Table;

import java.util.List;
import static org.junit.Assert.*;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;

/**
 *
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
@RunWith(JUnit4.class)
public class HierarchicalGridCVC3GeneratorTest {

    // Exercise the cvc3 generator using a table designed in matlab.
    @Test
    public void exerciseCVC3Generator() {
        Table table = new Table(null, variableCollection);

        final List<HierarchicalCell> topCells = table.getLeftGrid().getSubHierarchy();

        topCells.add(new HierarchicalCell("x > 0"));
        topCells.add(new HierarchicalCell("x == 0"));
        topCells.add(new HierarchicalCell("0 > x"));

        List<HierarchicalCell> nextGrid = topCells.get(1).getSubHierarchy();
        nextGrid.add(new HierarchicalCell("z > 0"));
        nextGrid.add(new HierarchicalCell("z == 0"));
        nextGrid.add(new HierarchicalCell("0 > z"));

        nextGrid.get(1).getSubHierarchy().add(new HierarchicalCell("z == 0"));

        String out = table.walk(new HierarchicalGridCVC3Generator(table.getVariables(), 1));

        String expected = "ECHO \"begin1\";\n"+
"PUSH;\n"+
"QUERY (NOT ( (x > 0) AND (x = 0) ) AND\n"+
"NOT ( (x > 0) AND (0 > x) ) AND\n"+
"NOT ( (x = 0) AND (0 > x) ));\n"+
"POP;\n"+
"ECHO \"end1\";\n"+
"ECHO \"begin2\";\n"+
"PUSH;\n"+
"QUERY ((x > 0) OR (x = 0) OR (0 > x));\n"+
"POP;\n"+
"ECHO \"end2\";\n"+
"ECHO \"begin3\";\n"+
"PUSH;\n"+
"QUERY ((x = 0)) => (NOT ( (z > 0) AND (z = 0) ) AND\n"+
"NOT ( (z > 0) AND (0 > z) ) AND\n"+
"NOT ( (z = 0) AND (0 > z) ));\n"+
"POP;\n"+
"ECHO \"end3\";\n"+
"ECHO \"begin4\";\n"+
"PUSH;\n"+
"QUERY ((x = 0)) => ((z > 0) OR (z = 0) OR (0 > z));\n"+
"POP;\n"+
"ECHO \"end4\";\n"+
"ECHO \"begin5\";\n"+
"PUSH;\n"+
"QUERY ((z = 0) AND (x = 0)) => ();\n"+
"POP;\n"+
"ECHO \"end5\";\n"+
"ECHO \"begin6\";\n"+
"PUSH;\n"+
"QUERY ((z = 0) AND (x = 0)) => ((z = 0));\n"+
"POP;\n"+
"ECHO \"end6\";\n";
        assertEquals(expected, out);
    }

    // Exercise the cvc3 generator's overflow protection
    @Test
    public void testCVC3GeneratorUnsignedOverflow() {
        Table table = new Table(null, new VariableCollection(variableParser.parseVariables("x:fixedt(0,8,0),y:fixedt(0,8,0)")));

        final List<HierarchicalCell> topCells = table.getLeftGrid().getSubHierarchy();

        topCells.add(new HierarchicalCell("(x + y) > 0"));
        topCells.add(new HierarchicalCell("(x * y) == 0"));
        topCells.add(new HierarchicalCell("(x / y) > 0"));

        List<HierarchicalCell> nextGrid = topCells.get(1).getSubHierarchy();
        nextGrid.add(new HierarchicalCell("(x - y) > 0"));

        String out = table.walk(new HierarchicalGridCVC3Generator(table.getVariables(), 1));

        String expected = "ECHO \"begin1\";\n" +
                "PUSH;\n" +
                "QUERY BVPLUS(9,x,y)[9:8] = 0bin0;\n" +
                "POP;\n" +
                "ECHO \"end1\";\n" +
                "ECHO \"begin2\";\n" +
                "PUSH;\n" +
                "QUERY BVMULT(16,x,y)[15:8] = 0bin00000000;\n" +
                "POP;\n" +
                "ECHO \"end2\";\n" +
                "ECHO \"begin3\";\n" +
                "PUSH;\n" +
                "QUERY (NOT ( BVGT(BVPLUS(8,x,y),0bin00000000) AND (BVMULT(8,x,y) = 0bin00000000) ) AND\n" +
                "NOT ( BVGT(BVPLUS(8,x,y),0bin00000000) AND BVGT(BVUDIV(x,y),0bin00000000) ) AND\n" +
                "NOT ( (BVMULT(8,x,y) = 0bin00000000) AND BVGT(BVUDIV(x,y),0bin00000000) ));\n" +
                "POP;\n" +
                "ECHO \"end3\";\n" +
                "ECHO \"begin4\";\n" +
                "PUSH;\n" +
                "QUERY (BVGT(BVPLUS(8,x,y),0bin00000000) OR (BVMULT(8,x,y) = 0bin00000000) OR BVGT(BVUDIV(x,y),0bin00000000));\n" +
                "POP;\n" +
                "ECHO \"end4\";\n" +
                "ECHO \"begin5\";\n" +
                "PUSH;\n" +
                "QUERY BVLE(y,x);\n" +
                "POP;\n" +
                "ECHO \"end5\";\n" +
                "ECHO \"begin6\";\n" +
                "PUSH;\n" +
                "QUERY ((BVMULT(8,x,y) = 0bin00000000)) => ();\n" +
                "POP;\n" +
                "ECHO \"end6\";\n" +
                "ECHO \"begin7\";\n" +
                "PUSH;\n" +
                "QUERY ((BVMULT(8,x,y) = 0bin00000000)) => (BVGT(BVSUB(8,x,y),0bin00000000));\n" +
                "POP;\n" +
                "ECHO \"end7\";\n";
        assertEquals(expected, out);
    }

    @Test
    public void testCVC3GeneratorSignedOverflow() {
        Table table = new Table(null, new VariableCollection(variableParser.parseVariables("x:fixedt(1,8,0),y:fixedt(1,8,0)")));

        final List<HierarchicalCell> topCells = table.getLeftGrid().getSubHierarchy();

        topCells.add(new HierarchicalCell("(x + y) > 0"));
        topCells.add(new HierarchicalCell("(x * y) == 0"));
        topCells.add(new HierarchicalCell("(x / y) > 0"));

        List<HierarchicalCell> nextGrid = topCells.get(1).getSubHierarchy();
        nextGrid.add(new HierarchicalCell("(x - y) > 0"));

        String out = table.walk(new HierarchicalGridCVC3Generator(table.getVariables(), 1));

        String expected = "ECHO \"begin1\";\n" +
                "PUSH;\n" +
                "QUERY ((x[7:7] = 0bin0 AND y[7:7] = 0bin0) => BVPLUS(8,x,y)[7:7] = 0bin0) AND ((x[7:7] = 0bin1 AND y[7:7] = 0bin1) => BVPLUS(8,x,y)[7:7] = 0bin1);\n" +
                "POP;\n" +
                "ECHO \"end1\";\n" +
                "ECHO \"begin2\";\n" +
                "PUSH;\n" +
                "QUERY ((x[7:7] = y[7:7]) => BVMULT(16,SX(x,16),SX(y,16))[15:7] = 0bin000000000) AND ((x[7:7] /= y[7:7]) => BVMULT(16,SX(x,16),SX(y,16))[15:7] = 0bin111111111);\n" +
                "POP;\n" +
                "ECHO \"end2\";\n" +
                "ECHO \"begin3\";\n" +
                "PUSH;\n" +
                "QUERY NOT (x = 0bin10000000 AND y = 0bin11111111);\n" +
                "POP;\n" +
                "ECHO \"end3\";\n" +
                "ECHO \"begin4\";\n" +
                "PUSH;\n" +
                "QUERY (NOT ( BVSGT(BVPLUS(8,x,y),0bin00000000) AND (BVMULT(8,x,y) = 0bin00000000) ) AND\n" +
                "NOT ( BVSGT(BVPLUS(8,x,y),0bin00000000) AND BVSGT(BVSDIV(x,y),0bin00000000) ) AND\n" +
                "NOT ( (BVMULT(8,x,y) = 0bin00000000) AND BVSGT(BVSDIV(x,y),0bin00000000) ));\n" +
                "POP;\n" +
                "ECHO \"end4\";\n" +
                "ECHO \"begin5\";\n" +
                "PUSH;\n" +
                "QUERY (BVSGT(BVPLUS(8,x,y),0bin00000000) OR (BVMULT(8,x,y) = 0bin00000000) OR BVSGT(BVSDIV(x,y),0bin00000000));\n" +
                "POP;\n" +
                "ECHO \"end5\";\n" +
                "ECHO \"begin6\";\n" +
                "PUSH;\n" +
                "QUERY ((x[7:7] = 0bin0 AND y[7:7] = 0bin1) => BVPLUS(8,x,y)[7:7] = 0bin0) AND ((x[7:7] = 0bin1 AND y[7:7] = 0bin0) => BVPLUS(8,x,y)[7:7] = 0bin1);\n" +
                "POP;\n" +
                "ECHO \"end6\";\n" +
                "ECHO \"begin7\";\n" +
                "PUSH;\n" +
                "QUERY ((BVMULT(8,x,y) = 0bin00000000)) => ();\n" +
                "POP;\n" +
                "ECHO \"end7\";\n" +
                "ECHO \"begin8\";\n" +
                "PUSH;\n" +
                "QUERY ((BVMULT(8,x,y) = 0bin00000000)) => (BVSGT(BVSUB(8,x,y),0bin00000000));\n" +
                "POP;\n" +
                "ECHO \"end8\";\n";
        assertEquals(expected, out);
    }

    VariableParser variableParser = new VariableParser();
    final VariableCollection variableCollection = new VariableCollection(variableParser.parseVariables("x,z"));
}
