/*
 * Copyright (C) 2014 Yanjun Jiang <jiangy76@mcmaster.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.eventbgenerator;

import ca.mcscert.jtet.expression.*;
import ca.mcscert.jtet.tablularexpression.*;
import ca.mcscert.jtet.parsers.VariableParser;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.*;
import static org.junit.Assert.assertEquals;

import java.io.*;
import java.util.*;

/**
 * @author Yanjun Jiang <jiangy76@mcmaster.ca>
 */
@RunWith(Parameterized.class)
public class ParameterizedEventBTableGeneratorTest {
    @Rule
    public TemporaryFolder folder = new TemporaryFolder();

    @Parameters
    public static Collection<Object[]> data() {
        return Arrays.asList(new Object[][]{
                {
                  makeOneLayerTable(),
                        1
                },
                {
                  makeTwoLayerTable(),
                        2
                },
                {
                  makeThreeLayerTable(),
                        3
                },
                {
                  makeIIPTable(),
                        1
                },
                {
                        makeMultipleOutputTable(),
                        3
                },
        });
    }

    @Parameter(value = 0)
    public Table m_table;

    @Parameter(value = 1)
    public int layerNum;

    @Test
    public void testNRTableEventBGenerator() throws IOException {
        m_folderPath = folder.getRoot().toString() + m_table.getTableName() + "/";
        doTest(RefinementMode.WithoutRefinement);
    }

    @Test
    public void testWRTableEventBGenerator() throws IOException {
        m_folderPath = folder.getRoot().toString() + m_table.getTableName() + "/";
        doTest(RefinementMode.WithRefinement);
    }

    public void setFileName(RefinementMode refinementMode) {
        m_generatedFileName = new ArrayList<String>(layerNum);
        m_expectedFileName  = new ArrayList<String>(layerNum);

        for (int i = 0; i < layerNum; i++) {
            m_generatedFileName.add(m_folderPath + generateFileName(refinementMode, i));
            m_expectedFileName.add(getExpectedFileName(refinementMode, i));
        }
    }

    public static Table makeOneLayerTable() {
        Variable outputVar = new Variable("output", new RealVariableType());
        Table table = initialize("oneLayerTable", "x", outputVar);
        HierarchicalGrid leftGrid = table.getLeftGrid();

        List<HierarchicalCell> cells = leftGrid.getSubHierarchy();
        cells.add(new HierarchicalCell("x >= 0"));
        cells.add(new HierarchicalCell("x < 0"));

        TwoDimensionalGrid outputGrid = new TwoDimensionalGrid(1, 2);
        outputGrid.get(0,0).setContents("1");
        outputGrid.get(0,1).setContents("2");
        table.getVariableOutputs().put(outputVar, outputGrid);

        return table;
    }

    public static Table makeTwoLayerTable() {
        Variable outputVar = new Variable("output", new RealVariableType());
        Table table = initialize("twoLayerTable", "x,z", outputVar);
        HierarchicalGrid leftGrid = table.getLeftGrid();

        List<HierarchicalCell> cells = leftGrid.getSubHierarchy();
        cells.add(new HierarchicalCell("x > 3-3"));
        cells.add(new HierarchicalCell("x == 0"));
        cells.add(new HierarchicalCell("x < 0"));

        List<HierarchicalCell> nextGrid = cells.get(1).getSubHierarchy();
        nextGrid.add(new HierarchicalCell("z > 0"));
        nextGrid.add(new HierarchicalCell("z == 0"));
        nextGrid.add(new HierarchicalCell("z < 0"));

        TwoDimensionalGrid outputGrid = new TwoDimensionalGrid(1, 5);
        outputGrid.get(0,0).setContents("1");
        outputGrid.get(0,1).setContents("1+1");
        outputGrid.get(0,2).setContents("3");
        outputGrid.get(0,3).setContents("4");
        outputGrid.get(0,4).setContents("5");
        table.getVariableOutputs().put(outputVar, outputGrid);

        return table;
    }

    public static Table makeThreeLayerTable() {
        Variable outputVar = new Variable("output", new RealVariableType());
        Table table = initialize("threeLayerTable", "x,y,z,w", outputVar);

        HierarchicalGrid leftGrid = table.getLeftGrid();

        List<HierarchicalCell> cells = leftGrid.getSubHierarchy();
        cells.add(new HierarchicalCell("x > 0"));
        cells.add(new HierarchicalCell("x == 0"));
        cells.add(new HierarchicalCell("x < 0"));

        List<HierarchicalCell> nextGrid = cells.get(0).getSubHierarchy();
        nextGrid.add(new HierarchicalCell("y >= 0"));
        nextGrid.add(new HierarchicalCell("y < 0"));
        nextGrid.get(1).getSubHierarchy().add(new HierarchicalCell("z > 0"));
        nextGrid.get(1).getSubHierarchy().add(new HierarchicalCell("z == 0"));
        nextGrid.get(1).getSubHierarchy().add(new HierarchicalCell("z < 0"));

        nextGrid = cells.get(2).getSubHierarchy();
        nextGrid.add(new HierarchicalCell("w > 1"));
        nextGrid.add(new HierarchicalCell("w == 1"));
        nextGrid.add(new HierarchicalCell("w < 1"));

        TwoDimensionalGrid outputGrid = new TwoDimensionalGrid(1, 8);
        outputGrid.get(0,0).setContents("1");
        outputGrid.get(0,1).setContents("2");
        outputGrid.get(0,2).setContents("3");
        outputGrid.get(0,3).setContents("4");
        outputGrid.get(0,4).setContents("5");
        outputGrid.get(0,5).setContents("6");
        outputGrid.get(0,6).setContents("7");
        outputGrid.get(0,7).setContents("8");
        table.getVariableOutputs().put(outputVar, outputGrid);

        return table;
    }

    public static Table makeMultipleOutputTable() {
        Variable outputVar1 = new Variable("output1", new RealVariableType());
        Variable outputVar2 = new Variable("output2", new RealVariableType());
        Table table = initialize("momThreeLayerTable", "x,y,z,w", outputVar1, outputVar2);

        HierarchicalGrid leftGrid = table.getLeftGrid();

        List<HierarchicalCell> cells = leftGrid.getSubHierarchy();
        cells.add(new HierarchicalCell("x > 0"));
        cells.add(new HierarchicalCell("x == 0"));
        cells.add(new HierarchicalCell("x < 0"));

        List<HierarchicalCell> nextGrid = cells.get(0).getSubHierarchy();
        nextGrid.add(new HierarchicalCell("y >= 0"));
        nextGrid.add(new HierarchicalCell("y < 0"));
        nextGrid.get(1).getSubHierarchy().add(new HierarchicalCell("z > 0"));
        nextGrid.get(1).getSubHierarchy().add(new HierarchicalCell("z == 0"));
        nextGrid.get(1).getSubHierarchy().add(new HierarchicalCell("z < 0"));

        nextGrid = cells.get(2).getSubHierarchy();
        nextGrid.add(new HierarchicalCell("w > 1"));
        nextGrid.add(new HierarchicalCell("w == 1"));
        nextGrid.add(new HierarchicalCell("w < 1"));

        TwoDimensionalGrid outputGrid = new TwoDimensionalGrid(1, 8);
        outputGrid.get(0,0).setContents("1");
        outputGrid.get(0,1).setContents("2");
        outputGrid.get(0,2).setContents("3");
        outputGrid.get(0,3).setContents("4");
        outputGrid.get(0,4).setContents("5");
        outputGrid.get(0,5).setContents("6");
        outputGrid.get(0,6).setContents("7");
        outputGrid.get(0,7).setContents("8");
        table.getVariableOutputs().put(outputVar1, outputGrid);

        outputGrid = new TwoDimensionalGrid(1, 8);
        outputGrid.get(0,0).setContents("9");
        outputGrid.get(0,1).setContents("10");
        outputGrid.get(0,2).setContents("11");
        outputGrid.get(0,3).setContents("12");
        outputGrid.get(0,4).setContents("13");
        outputGrid.get(0,5).setContents("14");
        outputGrid.get(0,6).setContents("15");
        outputGrid.get(0,7).setContents("16");
        table.getVariableOutputs().put(outputVar2, outputGrid);

        return table;
    }

    public static Table makeIIPTable() {
        //C_pwrConfirm is treated as enumerate here. It can choose values from
        //False:0 True:1
        Variable outputVar = new Variable("cPwrConfirm", new RealVariableType());

        Table table = initialize("IIP", "cPwrStatus", outputVar);
        HierarchicalGrid leftGrid = table.getLeftGrid();

        //C_pwrStatus is treated as enumerate here. It can choose values from
        //POST:0 NC:1 Ready:2 Standby:3 OffReq 4
        List<HierarchicalCell> cells = leftGrid.getSubHierarchy();
        cells.add(new HierarchicalCell("cPwrStatus == 4"));
        cells.add(new HierarchicalCell("cPwrStatus ~= 4"));

        TwoDimensionalGrid outputGrid = new TwoDimensionalGrid(1, 2);
        outputGrid.get(0,0).setContents("0");
        outputGrid.get(0,1).setContents("1");
        table.getVariableOutputs().put(outputVar, outputGrid);

        return table;
    }

    public static Table initialize(String tableName, String inputVariables, Variable...outputVariables) {
        Map<String, Variable> outputVariablesMap = new LinkedHashMap<String, Variable>(1);
        for(Variable outputVariable : outputVariables) {
            outputVariablesMap.put(outputVariable.name(), outputVariable);
        }

        VariableCollection variableCollection = new VariableCollection(m_variableParser.parseVariables(inputVariables), new PartialVariableCollection(outputVariablesMap));
        return new Table(tableName, variableCollection);
    }

    public void doTest(RefinementMode refineMode) throws IOException {
        setFileName(refineMode);
        execute(refineMode);
        assertResult(layerNum);
    }

    public String generateFileName(RefinementMode refinementMode, int refineLayerIndex) {
        if (refinementMode == RefinementMode.WithRefinement) {
            return m_table.getTableName() + "_M" + refineLayerIndex + ".bum";
        } else {
            return m_table.getTableName()  + ".bum";
        }
    }
    public String getExpectedFileName(RefinementMode refinementMode, int refineLayerIndex) {
        if (refinementMode == RefinementMode.WithRefinement){
            return "/expected/WR/" + m_table.getTableName() + "/" + m_table.getTableName() + "_M" + refineLayerIndex + ".bum";
        } else {
            return "/expected/NR/" + m_table.getTableName() + "/" + m_table.getTableName() + ".bum";
        }
    }

    private void execute(RefinementMode refineMode) throws IOException {
        EventBProject eventBProject = new EventBProject(m_table.getTableName());
        eventBProject.setTable(m_table);
        eventBProject.generate(m_folderPath, refineMode);
    }

    public void assertResult(int layerNum) throws IOException {
        for (int i = 0; i < layerNum; i++) {
            String generated, expected;
            InputStream inStream = null;
            try {
                inStream = new FileInputStream(m_generatedFileName.get(i));
                generated = readOutput(inStream);
            } finally {
                if (inStream != null) {
                    inStream.close();
                }
            }

        try {
            inStream = ParameterizedEventBTableGeneratorTest.class.getResourceAsStream(m_expectedFileName.get(i));
            expected = readOutput(inStream);
        } finally {
            if (inStream != null) {
                inStream.close();
            }
        }
            assertEquals(expected, generated);
        }
    }

    private String readOutput(InputStream inStream) throws IOException {
        String result = "";
        Reader reader = null;

        try {
            reader = new InputStreamReader(inStream);
            int character;
            while ((character = reader.read()) != -1) {
                result += (char) character;
            }
        } finally {
            if (reader != null) {
                reader.close();
            }
        }

        return result;
    }

    String m_folderPath;
    List<String>  m_generatedFileName;
    List<String>  m_expectedFileName;
    static VariableParser m_variableParser = new VariableParser();
}

