/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.expression;

import ca.mcscert.jtet.cvc3generator.CVC3VariablesDeclarationGenerator;
import ca.mcscert.jtet.eventbgenerator.EventBVariablesDeclarationGenerator;
import ca.mcscert.jtet.parsers.VariableParser;
import ca.mcscert.jtet.salgenerator.SALVariablesDeclarationGenerator;
import ca.mcscert.jtet.smtlibchecker.SMTLIBVariablesDeclarationGenerator;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;

import java.util.*;

import static org.junit.Assert.*;

/**
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
@RunWith(Parameterized.class)
public class GenericVariablesDeclarationGeneratorTest {
    @Parameters
    public static Collection<Object[]> data() {
        return Arrays.asList(new Object[][]{
                {new SMTLIBVariablesDeclarationGenerator(),
                        "(declare-fun x () Real)\n",
                        "(declare-fun new_x_name () Real)\n",
                        "(declare-fun x () Real)\n(assert (> x 0.0))\n",
                        "(declare-fun x () Bool)\n",
                        null,
                        null,
                        "(declare-fun x () MyEnum)\n" +
                                "(assert (isMyEnum x))\n",
                         "(declare-fun new_x_name () MyEnum)\n" +
                                "(assert (isMyEnum new_x_name))\n",
                         "(declare-fun x () MyEnum)\n" +
                                "(assert (isMyEnum x))\n" +
                                "(declare-fun y () MyEnum)\n" +
                                "(assert (isMyEnum y))\n"
                },
                {new CVC3VariablesDeclarationGenerator(),
                        "x:REAL;\n",
                        "new_x_name:REAL;\n",
                        "x:REAL;\nASSERT (x > 0);\n",
                        "x:BOOLEAN;\n",
                        "x:BITVECTOR(8);\n",
                        "x:BITVECTOR(8);\n",
                        "x:MyEnum;\n",
                        "new_x_name:MyEnum;\n",
                        "x:MyEnum;\ny:MyEnum;\n"
                },
                {new EventBVariablesDeclarationGenerator(),
                        "<org.eventb.core.variable de.prob.units.inferredUnitPragmaAttribute=\"\" de.prob.units.unitPragmaAttribute=\"\" name=\"x\" org.eventb.core.identifier=\"x\"/>\n" +
                        "<org.eventb.core.invariant name=\"Invariant1\" org.eventb.core.label=\"inv1\" org.eventb.core.predicate=\"x ∈ ℤ\"/>\n" +
                        "<org.eventb.core.event name=\"Event0\" org.eventb.core.convergence=\"0\" org.eventb.core.extended=\"false\" org.eventb.core.label=\"INITIALISATION\">\n" +
                        "<org.eventb.core.action name=\"InitAction1\" org.eventb.core.assignment=\"x :∈ ℤ\" org.eventb.core.label=\"act1\"/>\n" +
                        "</org.eventb.core.event>\n",
                        "<org.eventb.core.variable de.prob.units.inferredUnitPragmaAttribute=\"\" de.prob.units.unitPragmaAttribute=\"\" name=\"new_x_name\" org.eventb.core.identifier=\"new_x_name\"/>\n" +
                        "<org.eventb.core.invariant name=\"Invariant1\" org.eventb.core.label=\"inv1\" org.eventb.core.predicate=\"new_x_name ∈ ℤ\"/>\n" +
                        "<org.eventb.core.event name=\"Event0\" org.eventb.core.convergence=\"0\" org.eventb.core.extended=\"false\" org.eventb.core.label=\"INITIALISATION\">\n" +
                        "<org.eventb.core.action name=\"InitAction1\" org.eventb.core.assignment=\"new_x_name :∈ ℤ\" org.eventb.core.label=\"act1\"/>\n" +
                        "</org.eventb.core.event>\n",
                        null,
                        null,
                        null,
                        null,
                        null,
                        null,
                        null
                },
                {new SALVariablesDeclarationGenerator(),
                        "x:REAL\n",
                        "new_x_name:REAL\n",
                        "x:{x:REAL|(x > 0)}\n",
                        "x:BOOLEAN\n",
                        null,
                        null,
                        null,
                        null,
                        null
                },
        });
    }
    @Parameter(value = 0)
    public VariablesDeclarationGenerator generator;

    @Parameter(value = 1)
    public String RealExpectedOutput;

    @Parameter(value = 2)
    public String RealExpectedOutputWithNameChange;

    @Parameter(value = 3)
    public String RealWithPredicateSubtypeExpectedOutput;

    @Parameter(value = 4)
    public String BooleanExpectedOutput;

    @Parameter(value = 5)
    public String SignedFixedPointExpectedOutput;

    @Parameter(value = 6)
    public String UnsignedFixedPointExpectedOutput;

    @Parameter(value = 7)
    public String SingleVariableSingleEnumerationExpectedOutput;

    @Parameter(value = 8)
    public String SingleVariableSingleEnumerationExpectedWithNameChangeOutput;

    @Parameter(value = 9)
    public String MultipleVariableSingleEnumerationExpectedOutput;

    @Test
    public void RealTest() {
        if (RealExpectedOutput != null) {
            VariableCollection vars = new VariableCollection(variableParser.parseVariables("x"));
            assertEquals(RealExpectedOutput, generator.GenerateVariablesDeclaration(vars));
        }
    }

    @Test
    public void RealWithPredicateSubtypeTest() {
        if (RealWithPredicateSubtypeExpectedOutput != null) {
            VariableCollection vars = new VariableCollection(variableParser.parseVariables("x:{p:REAL|p > 0}"));
            assertEquals(RealWithPredicateSubtypeExpectedOutput, generator.GenerateVariablesDeclaration(vars));
        }
    }

    @Test
    public void RealWithNameChangeTest() {
        if (RealExpectedOutputWithNameChange != null) {
            VariableCollection vars = new VariableCollection(variableParser.parseVariables("x"));
            vars.getInputVariables().get("x").setOutputName("new_x_name");
            assertEquals(RealExpectedOutputWithNameChange, generator.GenerateVariablesDeclaration(vars));
        }
    }

    @Test
    public void BooleanTest() {
        if (BooleanExpectedOutput != null) {
            VariableCollection vars = new VariableCollection(variableParser.parseVariables("x:bool"));
            assertEquals(BooleanExpectedOutput, generator.GenerateVariablesDeclaration(vars));
        }
    }

    @Test
    public void SignedFixedPointTest() {
        if (SignedFixedPointExpectedOutput != null) {
            VariableCollection vars = new VariableCollection(variableParser.parseVariables("x:fixedt(1, 8, 2)"));
            assertEquals(SignedFixedPointExpectedOutput, generator.GenerateVariablesDeclaration(vars));
        }
    }

    @Test
    public void UnsignedFixedPointTest() {
        if (UnsignedFixedPointExpectedOutput != null) {
            VariableCollection vars = new VariableCollection(variableParser.parseVariables("x:fixedt(0, 8, 2)"));
            assertEquals(UnsignedFixedPointExpectedOutput, generator.GenerateVariablesDeclaration(vars));
        }
    }

    @Test
    public void SingleVariableSingleEnumerationTest() {
        if (SingleVariableSingleEnumerationExpectedOutput != null) {
            VariableCollection vars = new VariableCollection(variableParser.parseVariables("x:MyEnum"));
            assertEquals(SingleVariableSingleEnumerationExpectedOutput, generator.GenerateVariablesDeclaration(vars));
        }
    }

    @Test
    public void SingleVariableSingleEnumerationWithNameChangeTest() {
        if (SingleVariableSingleEnumerationExpectedWithNameChangeOutput != null) {
            VariableCollection vars = new VariableCollection(variableParser.parseVariables("x:MyEnum"));
            vars.getInputVariables().get("x").setOutputName("new_x_name");
            assertEquals(SingleVariableSingleEnumerationExpectedWithNameChangeOutput, generator.GenerateVariablesDeclaration(vars));
        }
    }

    // This ensures that the enumeration definition is only emitted once.  Other variable generations don't have this,
    // as they don't emit type information.
    @Test
    public void MultipleVariableSingleEnumerationTest() {
        if (MultipleVariableSingleEnumerationExpectedOutput != null) {
            VariableCollection vars = new VariableCollection(variableParser.parseVariables("x:MyEnum, y:MyEnum"));

            assertEquals(MultipleVariableSingleEnumerationExpectedOutput, generator.GenerateVariablesDeclaration(vars));
        }
    }

    @Before
    public void SetupVariableParser() {
        EnumerationVariableType myEnum = new EnumerationVariableType("MyEnum");
        myEnum.enumerationValues().add("e1");
        myEnum.enumerationValues().add("e2");
        myEnum.enumerationValues().add("e3");

        variableParser.addCustomType("MyEnum", myEnum);
    }

    VariableParser variableParser = new VariableParser();
}
